"""Utility functions for the language server."""

import asyncio
import builtins
import importlib.util
import os
import sys
from functools import wraps
from typing import Any, Awaitable, Callable, Coroutine, Optional, ParamSpec, TypeVar

import jaclang.compiler.absyntree as ast
from jaclang.compiler.codeloc import CodeLocInfo
from jaclang.compiler.symtable import Symbol, SymbolTable
from jaclang.utils.helpers import import_target_to_relative_path

import lsprotocol.types as lspt

T = TypeVar("T", bound=Callable[..., Coroutine[Any, Any, Any]])
P = ParamSpec("P")


def debounce(wait: float) -> Callable[[T], Callable[..., Awaitable[None]]]:
    """Debounce decorator for async functions."""

    def decorator(fn: T) -> Callable[..., Awaitable[None]]:
        @wraps(fn)
        async def debounced(*args: P.args, **kwargs: P.kwargs) -> None:
            async def call_it() -> None:
                await fn(*args, **kwargs)

            if hasattr(debounced, "_task"):
                debounced._task.cancel()

            async def debounced_coro() -> None:
                try:
                    await asyncio.sleep(wait)
                    await call_it()
                except asyncio.CancelledError:
                    pass

            setattr(  # noqa: B010
                debounced, "_task", asyncio.create_task(debounced_coro())
            )

        return debounced

    return decorator


def sym_tab_list(sym_tab: SymbolTable, file_path: str) -> list[SymbolTable]:
    """Iterate through symbol table."""
    sym_tabs = (
        [sym_tab]
        if not (
            isinstance(sym_tab.owner, ast.Module)
            and sym_tab.owner.loc.mod_path != file_path
        )
        else []
    )
    for i in sym_tab.kid:
        sym_tabs += sym_tab_list(i, file_path=file_path)
    return sym_tabs


def find_deepest_symbol_node_at_pos(
    node: ast.AstNode, line: int, character: int
) -> Optional[ast.AstSymbolNode]:
    """Return the deepest symbol node that contains the given position."""
    last_symbol_node = None

    if position_within_node(node, line, character):
        if isinstance(node, ast.AstSymbolNode):
            last_symbol_node = node

        for child in [i for i in node.kid if i.loc.mod_path == node.loc.mod_path]:
            if position_within_node(child, line, character):
                deeper_node = find_deepest_symbol_node_at_pos(child, line, character)
                if deeper_node is not None:
                    last_symbol_node = deeper_node

    return last_symbol_node


def position_within_node(node: ast.AstNode, line: int, character: int) -> bool:
    """Check if the position falls within the node's location."""
    if node.loc.first_line < line + 1 < node.loc.last_line:
        return True
    if (
        node.loc.first_line == line + 1
        and node.loc.col_start <= character + 1
        and (
            node.loc.last_line == line + 1
            and node.loc.col_end >= character + 1
            or node.loc.last_line > line + 1
        )
    ):
        return True
    if (
        node.loc.last_line == line + 1
        and node.loc.col_start <= character + 1 <= node.loc.col_end
    ):
        return True
    return False


def collect_symbols(node: SymbolTable) -> list[lspt.DocumentSymbol]:
    """Recursively collect symbols from the AST."""
    symbols = []
    if node is None:
        return symbols

    for key, item in node.tab.items():
        if (
            key in dir(builtins)
            or item in [owner_sym(tab) for tab in node.kid]
            or item.decl.loc.mod_path != node.owner.loc.mod_path
        ):
            continue
        else:

            pos = create_range(item.decl.loc)
            symbol = lspt.DocumentSymbol(
                name=key,
                kind=kind_map(item.decl),
                range=pos,
                selection_range=pos,
                children=[],
            )
            symbols.append(symbol)

    for sub_tab in [
        i for i in node.kid if i.owner.loc.mod_path == node.owner.loc.mod_path
    ]:
        sub_symbols = collect_symbols(sub_tab)

        if isinstance(
            sub_tab.owner,
            (ast.IfStmt, ast.ElseStmt, ast.WhileStmt, ast.IterForStmt, ast.InForStmt),
        ):
            symbols.extend(sub_symbols)
        else:
            sub_pos = create_range(sub_tab.owner.loc)
            symbol = lspt.DocumentSymbol(
                name=sub_tab.name,
                kind=kind_map(sub_tab.owner),
                range=sub_pos,
                selection_range=sub_pos,
                children=sub_symbols,
            )
            symbols.append(symbol)

    return symbols


def owner_sym(table: SymbolTable) -> Optional[Symbol]:
    """Get owner sym."""
    if table.parent and isinstance(table.owner, ast.AstSymbolNode):
        return table.parent.lookup(table.owner.sym_name)
    return None


def create_range(loc: CodeLocInfo) -> lspt.Range:
    """Create an lspt.Range from a location object."""
    return lspt.Range(
        start=lspt.Position(
            line=loc.first_line - 1 if loc.first_line > 0 else 0,
            character=loc.col_start - 1 if loc.col_start > 0 else 0,
        ),
        end=lspt.Position(
            line=loc.last_line - 1 if loc.last_line > 0 else 0,
            character=loc.col_end - 1 if loc.col_end > 0 else 0,
        ),
    )


def kind_map(sub_tab: ast.AstNode) -> lspt.SymbolKind:
    """Map the symbol node to an lspt.SymbolKind."""
    return (
        lspt.SymbolKind.Function
        if isinstance(sub_tab, (ast.Ability, ast.AbilityDef))
        else (
            lspt.SymbolKind.Class
            if isinstance(sub_tab, (ast.Architype, ast.ArchDef))
            else (
                lspt.SymbolKind.Module
                if isinstance(sub_tab, ast.Module)
                else (
                    lspt.SymbolKind.Enum
                    if isinstance(sub_tab, (ast.Enum, ast.EnumDef))
                    else lspt.SymbolKind.Variable
                )
            )
        )
    )


def get_mod_path(mod_path: ast.ModulePath, name_node: ast.Name) -> str | None:
    """Get path for a module import name."""
    ret_target = None
    module_location_path = mod_path.loc.mod_path
    if mod_path.parent and (
        (
            isinstance(mod_path.parent.parent, ast.Import)
            and mod_path.parent.parent.hint.tag.value == "py"
        )
        or (
            isinstance(mod_path.parent, ast.Import)
            and mod_path.parent.from_loc
            and mod_path.parent.hint.tag.value == "py"
        )
    ):
        if mod_path.path and name_node in mod_path.path:
            temporary_path_str = ("." * mod_path.level) + ".".join(
                [p.value for p in mod_path.path[: mod_path.path.index(name_node) + 1]]
                if mod_path.path
                else ""
            )
        else:
            temporary_path_str = mod_path.path_str
        sys.path.append(os.path.dirname(module_location_path))
        spec = importlib.util.find_spec(temporary_path_str)
        sys.path.remove(os.path.dirname(module_location_path))
        if spec and spec.origin and spec.origin.endswith(".py"):
            ret_target = spec.origin
    elif mod_path.parent and (
        (
            isinstance(mod_path.parent.parent, ast.Import)
            and mod_path.parent.parent.hint.tag.value == "jac"
        )
        or (
            isinstance(mod_path.parent, ast.Import)
            and mod_path.parent.from_loc
            and mod_path.parent.hint.tag.value == "jac"
        )
    ):
        ret_target = import_target_to_relative_path(
            level=mod_path.level,
            target=mod_path.path_str,
            base_path=os.path.dirname(module_location_path),
        )
    return ret_target


def get_item_path(mod_item: ast.ModuleItem) -> tuple[str, tuple[int, int]] | None:
    """Get path."""
    item_name = mod_item.name.value
    if mod_item.from_parent.hint.tag.value == "py" and mod_item.from_parent.from_loc:
        path = get_mod_path(mod_item.from_parent.from_loc, mod_item.name)
        if path:
            return get_definition_range(path, item_name)
    elif mod_item.from_parent.hint.tag.value == "jac":
        mod_node = mod_item.from_mod_path
        if mod_node.sub_module and mod_node.sub_module._sym_tab:
            for symbol_name, symbol in mod_node.sub_module._sym_tab.tab.items():
                if symbol_name == item_name:
                    return symbol.decl.loc.mod_path, (
                        symbol.decl.loc.first_line - 1,
                        symbol.decl.loc.last_line - 1,
                    )
    return None


def get_definition_range(
    filename: str, name: str
) -> tuple[str, tuple[int, int]] | None:
    """Get the start and end line of a function or class definition in a file."""
    import ast

    with open(filename, "r") as file:
        source = file.read()

    tree = ast.parse(source)

    for node in ast.walk(tree):
        if isinstance(node, (ast.FunctionDef, ast.ClassDef)) and node.name == name:
            start_line = node.lineno
            end_line = (
                node.body[-1].end_lineno
                if hasattr(node.body[-1], "end_lineno")
                else node.body[-1].lineno
            )
            if start_line and end_line:
                return filename, (start_line - 1, end_line - 1)
        elif isinstance(node, ast.Assign):
            for target in node.targets:
                if isinstance(target, ast.Name) and target.id == name:
                    start_line = node.lineno
                    end_line = (
                        node.end_lineno if hasattr(node, "end_lineno") else node.lineno
                    )
                    if start_line and end_line:
                        return filename, (start_line - 1, end_line - 1)

    return None



"""Test case utils for Jaseci."""

import inspect
import os
from typing import Callable, Optional
from unittest import TestCase as _TestCase


import jaclang
from jaclang.compiler.passes import Pass
from jaclang.utils.helpers import get_ast_nodes_as_snake_case as ast_snakes


class TestCase(_TestCase):
    """Base test case for Jaseci."""

    def setUp(self) -> None:
        """Set up test case."""
        return super().setUp()

    def tearDown(self) -> None:
        """Tear down test case."""
        return super().tearDown()

    def load_fixture(self, fixture: str) -> str:
        """Load fixture from fixtures directory."""
        frame = inspect.currentframe()
        if frame is None or frame.f_back is None:
            raise ValueError("Unable to get the previous stack frame.")
        module = inspect.getmodule(frame.f_back)
        if module is None or module.__file__ is None:
            raise ValueError("Unable to determine the file of the module.")
        fixture_src = module.__file__
        fixture_path = os.path.join(os.path.dirname(fixture_src), "fixtures", fixture)
        with open(fixture_path, "r") as f:
            return f.read()

    def file_to_str(self, file_path: str) -> str:
        """Load fixture from fixtures directory."""
        with open(file_path, "r") as f:
            return f.read()

    def fixture_abs_path(self, fixture: str) -> str:
        """Get absolute path of a fixture from fixtures directory."""
        frame = inspect.currentframe()
        if frame is None or frame.f_back is None:
            raise ValueError("Unable to get the previous stack frame.")
        module = inspect.getmodule(frame.f_back)
        if module is None or module.__file__ is None:
            raise ValueError("Unable to determine the file of the module.")
        fixture_src = module.__file__
        file_path = os.path.join(os.path.dirname(fixture_src), "fixtures", fixture)
        return os.path.abspath(file_path)


class TestCaseMicroSuite(TestCase):
    """Base test case for Jaseci."""

    test_micro_jac_files_fully_tested: Optional[Callable[[TestCase], None]] = None
    methods: list[str] = []

    @classmethod
    def self_attach_micro_tests(cls) -> None:
        """Attach micro tests."""
        for filename in [
            os.path.normpath(os.path.join(root, name))
            for root, _, files in os.walk(
                os.path.dirname(os.path.dirname(jaclang.__file__))
            )
            for name in files
            if name.endswith(".jac") and "err" not in name
        ]:
            method_name = (
                f"test_micro_{filename.replace('.jac', '').replace(os.sep, '_')}"
            )
            cls.methods.append(method_name)
            setattr(cls, method_name, lambda self, f=filename: self.micro_suite_test(f))

        def test_micro_jac_files_fully_tested(self: TestCase) -> None:  # noqa: ANN001
            """Test that all micro jac files are fully tested."""
            for filename in cls.methods:
                if os.path.isfile(filename):
                    method_name = f"test_micro_{filename.replace('.jac', '').replace(os.sep, '_')}"
                    self.assertIn(method_name, dir(self))

        cls.test_micro_jac_files_fully_tested = test_micro_jac_files_fully_tested

    def micro_suite_test(self, filename: str) -> None:
        """Test micro jac file."""
        pass


class AstSyncTestMixin:
    """Mixin for testing AST sync."""

    TargetPass: Optional[Pass] = None

    def test_pass_ast_complete(self) -> None:
        """Test for enter/exit name diffs with parser."""
        ast_func_names = [
            x
            for x in ast_snakes()
            if x
            not in [
                "ast_node",
                "walker_stmt_only_node",
                "jac_source",
                "empty_token",
                "ast_symbol_node",
                "ast_symbol_stub_node",
                "ast_impl_needing_node",
                "ast_access_node",
                "token_symbol",
                "literal",
                "ast_doc_node",
                "ast_sem_str_node",
                "python_module_ast",
                "ast_async_node",
                "ast_else_body_node",
                "ast_typed_var_node",
                "ast_impl_only_node",
                "expr",
                "atom_expr",
                "element_stmt",
                "arch_block_stmt",
                "enum_block_stmt",
                "code_block_stmt",
                "name_atom",
                "arch_spec",
                "match_pattern",
            ]
        ]
        pygen_func_names = []
        for name, value in inspect.getmembers(self.TargetPass):
            if (
                (name.startswith("enter_") or name.startswith("exit_"))
                and inspect.isfunction(value)
                and not getattr(self.TargetPass.__base__, value.__name__, False)  # type: ignore
                and value.__qualname__.split(".")[0]
                == self.TargetPass.__name__.replace("enter_", "").replace("exit_", "")  # type: ignore
            ):
                pygen_func_names.append(name.replace("enter_", "").replace("exit_", ""))
        for name in pygen_func_names:
            self.assertIn(name, ast_func_names)  # type: ignore
        for name in ast_func_names:
            self.assertIn(name, pygen_func_names)  # type: ignore



"""The Jac Programming Language."""

from jaclang.plugin.default import (  # noqa: E402
    JacBuiltin,
    JacCmdDefaults,
    JacFeatureDefaults,
)
from jaclang.plugin.feature import JacFeature, pm  # noqa: E402

jac_import = JacFeature.jac_import

pm.register(JacFeatureDefaults)
pm.register(JacBuiltin)
pm.register(JacCmdDefaults)
pm.load_setuptools_entrypoints("jac")

__all__ = ["jac_import"]



"""Jaclang Language Server."""

from __future__ import annotations

from typing import Optional

from jaclang.compiler.constant import (
    JacSemTokenModifier as SemTokMod,
    JacSemTokenType as SemTokType,
)
from jaclang.langserve.engine import JacLangServer
from jaclang.langserve.utils import debounce

import lsprotocol.types as lspt

server = JacLangServer()


@server.feature(lspt.TEXT_DOCUMENT_DID_OPEN)
@server.feature(lspt.TEXT_DOCUMENT_DID_SAVE)
async def did_open(ls: JacLangServer, params: lspt.DidOpenTextDocumentParams) -> None:
    """Check syntax on change."""
    await ls.analyze_and_publish(params.text_document.uri)


@server.feature(lspt.TEXT_DOCUMENT_DID_CHANGE)
@debounce(0.3)
async def did_change(ls: JacLangServer, params: lspt.DidOpenTextDocumentParams) -> None:
    """Check syntax on change."""
    await ls.analyze_and_publish(params.text_document.uri, level=1)


@server.feature(lspt.TEXT_DOCUMENT_FORMATTING)
def formatting(
    ls: JacLangServer, params: lspt.DocumentFormattingParams
) -> list[lspt.TextEdit]:
    """Format the given document."""
    return ls.formatted_jac(params.text_document.uri)


@server.feature(
    lspt.WORKSPACE_DID_CREATE_FILES,
    lspt.FileOperationRegistrationOptions(
        filters=[
            lspt.FileOperationFilter(pattern=lspt.FileOperationPattern("**/*.jac"))
        ]
    ),
)
def did_create_files(ls: JacLangServer, params: lspt.CreateFilesParams) -> None:
    """Check syntax on file creation."""


@server.feature(
    lspt.WORKSPACE_DID_RENAME_FILES,
    lspt.FileOperationRegistrationOptions(
        filters=[
            lspt.FileOperationFilter(pattern=lspt.FileOperationPattern("**/*.jac"))
        ]
    ),
)
def did_rename_files(ls: JacLangServer, params: lspt.RenameFilesParams) -> None:
    """Check syntax on file rename."""
    new_uris = [file.new_uri for file in params.files]
    old_uris = [file.old_uri for file in params.files]
    for i in range(len(new_uris)):
        ls.rename_module(old_uris[i], new_uris[i])


@server.feature(
    lspt.WORKSPACE_DID_DELETE_FILES,
    lspt.FileOperationRegistrationOptions(
        filters=[
            lspt.FileOperationFilter(pattern=lspt.FileOperationPattern("**/*.jac"))
        ]
    ),
)
def did_delete_files(ls: JacLangServer, params: lspt.DeleteFilesParams) -> None:
    """Check syntax on file delete."""
    for file in params.files:
        ls.delete_module(file.uri)


@server.feature(
    lspt.TEXT_DOCUMENT_COMPLETION,
    lspt.CompletionOptions(trigger_characters=[".", ":", ""]),
)
def completion(ls: JacLangServer, params: lspt.CompletionParams) -> lspt.CompletionList:
    """Provide completion."""
    return ls.get_completion(params.text_document.uri, params.position)


@server.feature(lspt.TEXT_DOCUMENT_HOVER, lspt.HoverOptions(work_done_progress=True))
def hover(
    ls: JacLangServer, params: lspt.TextDocumentPositionParams
) -> Optional[lspt.Hover]:
    """Provide hover information for the given hover request."""
    return ls.get_hover_info(params.text_document.uri, params.position)


@server.feature(lspt.TEXT_DOCUMENT_DOCUMENT_SYMBOL)
def document_symbol(
    ls: JacLangServer, params: lspt.DocumentSymbolParams
) -> list[lspt.DocumentSymbol]:
    """Provide document symbols."""
    return ls.get_document_symbols(params.text_document.uri)


@server.feature(lspt.TEXT_DOCUMENT_DEFINITION)
def definition(
    ls: JacLangServer, params: lspt.TextDocumentPositionParams
) -> Optional[lspt.Location]:
    """Provide definition."""
    return ls.get_definition(params.text_document.uri, params.position)


@server.feature(
    lspt.TEXT_DOCUMENT_SEMANTIC_TOKENS_FULL,
    lspt.SemanticTokensLegend(
        token_types=SemTokType.as_str_list(),
        token_modifiers=SemTokMod.as_str_list(),
    ),
)
def semantic_tokens_full(
    ls: JacLangServer, params: lspt.SemanticTokensParams
) -> lspt.SemanticTokens:
    """Provide semantic tokens."""
    return ls.get_semantic_tokens(params.text_document.uri)


def run_lang_server() -> None:
    """Run the language server."""
    server.start_io()


if __name__ == "__main__":
    run_lang_server()



"""Test Jac reference examples."""

import io
import os
from contextlib import redirect_stdout
from typing import Callable, Optional

import jaclang
from jaclang.compiler.compile import jac_file_to_pass
from jaclang.plugin.feature import JacFeature as Jac
from jaclang.utils.test import TestCase


class JacReferenceTests(TestCase):
    """Test Reference examples."""

    test_ref_jac_files_fully_tested: Optional[Callable[[TestCase], None]] = None
    methods: list[str] = []

    @classmethod
    def self_attach_ref_tests(cls) -> None:
        """Attach micro tests."""
        for filename in [
            os.path.normpath(os.path.join(root, name))
            for root, _, files in os.walk(
                os.path.join(
                    os.path.dirname(os.path.dirname(jaclang.__file__)),
                    "examples/reference",
                )
            )
            for name in files
            if name.endswith(".jac") and not name.startswith("err")
        ]:
            method_name = (
                f"test_ref_{filename.replace('.jac', '').replace(os.sep, '_')}"
            )
            cls.methods.append(method_name)
            setattr(cls, method_name, lambda self, f=filename: self.micro_suite_test(f))

        def test_ref_jac_files_fully_tested(self: TestCase) -> None:  # noqa: ANN001
            """Test that all micro jac files are fully tested."""
            for filename in cls.methods:
                if os.path.isfile(filename):
                    method_name = (
                        f"test_ref_{filename.replace('.jac', '').replace(os.sep, '_')}"
                    )
                    self.assertIn(method_name, dir(self))

        cls.test_ref_jac_files_fully_tested = test_ref_jac_files_fully_tested

    def micro_suite_test(self, filename: str) -> None:
        """Test file."""

        def execute_and_capture_output(code: str | bytes, filename: str = "") -> str:
            Jac.get_root().reset()
            f = io.StringIO()
            with redirect_stdout(f):
                exec(
                    code,
                    {
                        "__file__": filename,
                        "__name__": "__main__",
                        "__jac_mod_bundle__": None,
                    },
                )
            return f.getvalue()

        try:
            if "tests.jac" in filename or "check_statements.jac" in filename:
                return
            jacast = jac_file_to_pass(filename).ir
            code_content = compile(
                source=jacast.gen.py_ast[0],
                filename=jacast.loc.mod_path,
                mode="exec",
            )
            output_jac = execute_and_capture_output(code_content, filename=filename)

            filename = filename.replace(".jac", ".py")
            with open(filename, "r") as file:
                code_content = file.read()
            output_py = execute_and_capture_output(code_content, filename=filename)

            # print(f"\nJAC Output:\n{output_jac}")
            # print(f"\nPython Output:\n{output_py}")

            self.assertGreater(len(output_py), 0)
            self.assertEqual(output_py, output_jac)
        except Exception as e:
            # print(f"\nJAC Output:\n{output_jac}")
            # print(f"\nPython Output:\n{output_py}")
            raise e


JacReferenceTests.self_attach_ref_tests()



"""Test Jac cli module."""

import inspect
import io
import os
import subprocess
import sys

from jaclang.cli import cli
from jaclang.plugin.builtin import dotgen
from jaclang.utils.test import TestCase


class JacCliTests(TestCase):
    """Test pass module."""

    def setUp(self) -> None:
        """Set up test."""
        return super().setUp()

    def test_jac_cli_run(self) -> None:
        """Basic test for pass."""
        captured_output = io.StringIO()
        sys.stdout = captured_output

        # Execute the function
        cli.run(self.fixture_abs_path("hello.jac"))

        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()

        self.assertIn("Hello World!", stdout_value)

    def test_jac_cli_alert_based_err(self) -> None:
        """Basic test for pass."""
        captured_output = io.StringIO()
        sys.stdout = captured_output
        sys.stderr = captured_output

        # Execute the function
        # try:
        cli.enter(self.fixture_abs_path("err2.jac"), entrypoint="speak", args=[])  # type: ignore
        # except Exception as e:
        #     print(f"Error: {e}")

        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__
        stdout_value = captured_output.getvalue()
        # print(stdout_value)
        self.assertIn("Errors occurred", stdout_value)

    def test_jac_ast_tool_pass_template(self) -> None:
        """Basic test for pass."""
        captured_output = io.StringIO()
        sys.stdout = captured_output

        cli.tool("pass_template")

        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        self.assertIn("Sub objects.", stdout_value)
        self.assertGreater(stdout_value.count("def exit_"), 10)

    def test_jac_cmd_line(self) -> None:
        """Basic test for pass."""
        process = subprocess.Popen(
            ["jac"],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        stdout_value, _ = process.communicate(input="exit\n")
        self.assertEqual(process.returncode, 0, "Process did not exit successfully")
        self.assertIn("Welcome to the Jac CLI!", stdout_value)

    def test_ast_print(self) -> None:
        """Testing for print AstTool."""
        captured_output = io.StringIO()
        sys.stdout = captured_output

        cli.tool("ir", ["ast", f"{self.fixture_abs_path('hello.jac')}"])

        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        self.assertIn("+-- Token", stdout_value)

    def test_ast_dotgen(self) -> None:
        """Testing for print AstTool."""
        captured_output = io.StringIO()
        sys.stdout = captured_output

        cli.tool("ir", ["ast.", f"{self.fixture_abs_path('hello.jac')}"])

        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        self.assertIn('[label="MultiString"]', stdout_value)

    def test_type_check(self) -> None:
        """Testing for print AstTool."""
        captured_output = io.StringIO()
        sys.stdout = captured_output
        cli.check(f"{self.fixture_abs_path('game1.jac')}")
        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        self.assertIn("Errors: 0, Warnings: 1", stdout_value)

    def test_type_info(self) -> None:
        """Testing for type info inside the ast tool."""
        captured_output = io.StringIO()
        sys.stdout = captured_output
        cli.tool("ir", ["ast", f"{self.fixture_abs_path('type_info.jac')}"])
        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        self.assertEqual(stdout_value.count("type_info.ServerWrapper"), 7)
        self.assertEqual(stdout_value.count("builtins.int"), 2)
        self.assertEqual(stdout_value.count("builtins.str"), 7)

    def test_build_and_run(self) -> None:
        """Testing for print AstTool."""
        if os.path.exists(f"{self.fixture_abs_path('needs_import.jir')}"):
            os.remove(f"{self.fixture_abs_path('needs_import.jir')}")
        captured_output = io.StringIO()
        sys.stdout = captured_output
        cli.build(f"{self.fixture_abs_path('needs_import.jac')}")
        cli.run(f"{self.fixture_abs_path('needs_import.jir')}")
        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        self.assertIn("Errors: 0, Warnings: 0", stdout_value)
        self.assertIn("<module 'pyfunc' from", stdout_value)

    def test_cache_no_cache_on_run(self) -> None:
        """Basic test for pass."""
        process = subprocess.Popen(
            ["jac", "run", f"{self.fixture_abs_path('hello_nc.jac')}", "-nc"],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        stdout, _ = process.communicate()
        self.assertFalse(
            os.path.exists(
                f"{self.fixture_abs_path(os.path.join('__jac_gen__', 'hello_nc.jbc'))}"
            )
        )
        self.assertIn("Hello World!", stdout)
        process = subprocess.Popen(
            ["jac", "run", f"{self.fixture_abs_path('hello_nc.jac')}", "-c"],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        stdout, _ = process.communicate()
        self.assertIn("Hello World!", stdout)
        self.assertTrue(
            os.path.exists(
                f"{self.fixture_abs_path(os.path.join('__jac_gen__', 'hello_nc.jbc'))}"
            )
        )
        os.remove(
            f"{self.fixture_abs_path(os.path.join('__jac_gen__', 'hello_nc.jbc'))}"
        )

    def test_run_test(self) -> None:
        """Basic test for pass."""
        process = subprocess.Popen(
            ["jac", "test", f"{self.fixture_abs_path('run_test.jac')}", "-m 2"],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        stdout, stderr = process.communicate()
        self.assertIn("Ran 3 tests", stderr)
        self.assertIn("FAILED (failures=2)", stderr)
        self.assertIn("F.F", stderr)

        process = subprocess.Popen(
            [
                "jac",
                "test",
                "-d" + f"{self.fixture_abs_path('../../../')}",
                "-f" + "circle*",
                "-x",
            ],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        stdout, stderr = process.communicate()
        self.assertIn("circle", stdout)
        self.assertNotIn("circle_purfe.test", stdout)
        self.assertNotIn("circle_pure.impl", stdout)

        process = subprocess.Popen(
            ["jac", "test", "-f" + "*run_test.jac", "-m 3"],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        stdout, stderr = process.communicate()
        self.assertIn("...F", stderr)
        self.assertIn("F.F", stderr)

    def test_graph_coverage(self) -> None:
        """Test for coverage of graph cmd."""
        graph_params = set(inspect.signature(cli.dot).parameters.keys())
        dotgen_params = set(inspect.signature(dotgen).parameters.keys())
        dotgen_params = dotgen_params - {"node", "dot_file", "edge_type"}
        dotgen_params.update({"initial", "saveto", "connection", "session"})
        self.assertTrue(dotgen_params.issubset(graph_params))
        self.assertEqual(len(dotgen_params) + 1, len(graph_params))

    def test_graph(self) -> None:
        """Test for graph CLI cmd."""
        captured_output = io.StringIO()
        sys.stdout = captured_output
        cli.dot(
            f"{self.fixture_abs_path('../../../examples/reference/connect_expressions.jac')}"
        )
        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        if os.path.exists("connect_expressions.dot"):
            os.remove("connect_expressions.dot")
        self.assertIn("11\n13\n15\n>>> Graph content saved to", stdout_value)
        self.assertIn("connect_expressions.dot\n", stdout_value)

    def test_py_to_jac(self) -> None:
        """Test for graph CLI cmd."""
        captured_output = io.StringIO()
        sys.stdout = captured_output
        cli.py2jac(f"{self.fixture_abs_path('../../tests/fixtures/pyfunc.py')}")
        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        self.assertIn("can my_print(x: object) -> None", stdout_value)



# flake8: noqa

"""Python function for testing py imports."""

from enum import Enum


class ShapeType(Enum):
    """Enum for shape types"""

    CIRCLE = "Circle"
    UNKNOWN = "Unknown"
    print("hello")


f: int = 34
a = f > 3
qq = {}
e = "hello"
ff = b"hello"


def my_print(x: object) -> None:
    """Print function."""
    print(x)


print("Hello world!")


class MyClass2:
    """This is a docstring for MyClass."""

    def __init__(self):
        """Constructor docstring."""
        pass

    def my_method(self) -> None:
        """Method docstring."""
        print("My method")
        return "done"


name, age, city = "Alice", 30, "Wonderland"
message = f"Hello, my name is {name}, I am {age} years old, and I live in {city}."
result = f"The result of 5 times 7 is {5 * 7}."


def my_some_func(a: int, b: int) -> int:
    pass


class MyClass:
    """My class"""

    def __init__(self, x):
        self.x = x

    def my_method(self):
        print("My method")


class Student(MyClass):
    """Student class"""

    def __init__(self, x, y):
        super().__init__(x)
        self.y = y

    def my_method(self):
        print("My method")


a = 9
for i in range(a):
    print(i)
    b = i
    if i == 5:
        break
    print(b)
else:
    print("Loop completed normally{}".format(i))


def fooo45() -> None:
    """Test func  fooo45"""
    pass


numbers = [1, 2, 3, 4, 5]
squares = [x**2 for x in numbers]
squares_dict = {x: x**2 for x in numbers}
squares_generator = (x**2 for x in numbers)
add = lambda x, y: x + y
numbers2 = [1, 2, 3, 3, 4, 5, 5]
unique_numbers2 = {x for x in numbers2}
even_numbers = {x for x in range(10) if x % 2 == 0}
coordinates = {(x, y) for x in range(3) for y in range(3)}
even_numbers4 = [num for num in range(11) if num % 2 == 0]
sentence = "the quick brown fox jumps over the lazy dog"
capitalized_words = [word.upper() for word in sentence.split() if len(word) > 3]
result78 = [
    (x, y) for x in range(1, 4) if x % 2 == 0 for y in range(1, 4) if y % 2 == 0
]
print(result78)


class Car:
    wheels: int = 4

    def __init__(self, make: str, model: str, year: int):
        self.make = make
        self.model = model
        self.year = year

    def display_car_info(self):
        print(f"Car Info: {self.year} {self.make} {self.model}")

    @staticmethod
    def get_wheels():
        return Car.wheels


car1 = Car("Toyota", "Camry", 2020)
car1.display_car_info()
print("Number of wheels:", Car.get_wheels())

p = print

p("Multiply:", 7 * 2)
p("Division:", 15 / 3)
p("Floor:", 15 // 3)
p("Modulo:", 17 % 5)
p("Expon:", 2**3)
p("combo:", (9 + 2) * 9 - 2)


for i in range(a):
    print(i)
    if i == 5:
        break
    print(b)
else:
    print("Loop completed normally{}".format(i))

numbers = [1, 2, 3, 4, 5]
squares = [x**2 for x in numbers]
squares_dict = {x: x**2 for x in numbers}
squares_generator = (x**2 for x in numbers)
add = lambda x, y: x + y  # noqa


def my_decorator(func):
    """Decorator"""

    def wrapper():
        print("Something is happening before the function is called.")
        func()
        print("Something is happening after the function is called.")

    return wrapper


@my_decorator
def say_hello():
    """Say hello"""
    print("Hello!")


def print_base_classes(cls: type) -> type:
    print(f"Base classes of {cls.__name__}: {[c.__name__ for c in cls.__bases__]}")
    return cls


def foo(cls):
    return cls


class Animal:
    pass


class domestic:
    pass


@print_base_classes
class Mammal(Animal, domestic):
    pass


s = []
a = [1, 2, 3, 4, 5]
b = (1, 2, 3, 4, 5)
x = ()
c = {1, 2, 3, 4, 5}
d = {1: "one", 2: "two", 3: "three", 4: "four", 5: "five"}


x = {}
x = {"name": "John", "age": 30, "city": "New York"}

x = 5
while x < 10:
    x += 1
    print("hello")

try:
    result = 10 / 0
except ZeroDivisionError:
    print("Cannot divide by zero!")
else:
    print(result)
finally:
    print("This will always execute.")


def foox():
    return 9 + 4


a, b, c = 1, 2, 3
x = y = z = 0
# first, *rest = [1, 2, 3, 4, 5]


def greet(name, greeting="Hello") -> None:
    """Greet someone."""
    print(f"{greeting}, {name}")


greet("Alice")
greet("Bob", greeting="Hi")


def average(*args):
    """Average function"""
    return sum(args) / len(args)


avg = average(1, 2, 3, 4, 5)


def greet2(**kwargs):
    """Greet someone."""
    print(f"Hello, {kwargs['name']}!")


greet2(name="Alice", age=30)

name = "Alice"
age = 30
print("Name: {}, Age: {}".format(name, age))


x = True
y = False
z = x and y


x = 5
assert x == 5, "x should be equal to 5"

uu = 9
if uu == 9:
    print("uu is 9")
elif uu == 10:
    print("uu is 10")
else:
    print("uu is not 9 or 10")


def my_functionx():
    """My function"""

    def my_inner_function():
        """Inner function"""
        return 9

    my_inner_function()
    for i in range(10):
        if i == 5:
            break
        if i == 3:
            continue
    pass


x = 90
if not isinstance(x, int):
    raise ValueError("x must be an integer")
else:
    print(x)


my_list = [1, 2, 3, 4, 5]
del my_list[2]
print(my_list)


x = 5
y = 10
z = 15

if not x == y:
    print("x is not equal to y")

if x == y or y == z:
    print("Either x is equal to y or y is equal to z")
else:
    print("Neither x is equal to y nor y is equal to z")



"""Living Workspace of Jac project."""

from __future__ import annotations

import asyncio
import logging
from concurrent.futures import ThreadPoolExecutor
from enum import IntEnum
from typing import Optional


import jaclang.compiler.absyntree as ast
from jaclang.compiler.compile import jac_ir_to_pass, jac_str_to_pass
from jaclang.compiler.parser import JacParser
from jaclang.compiler.passes import Pass
from jaclang.compiler.passes.main.schedules import type_checker_sched
from jaclang.compiler.passes.tool import FuseCommentsPass, JacFormatPass
from jaclang.compiler.passes.transform import Alert
from jaclang.langserve.utils import (
    collect_symbols,
    create_range,
    find_deepest_symbol_node_at_pos,
    get_item_path,
    get_mod_path,
)
from jaclang.vendor.pygls import uris
from jaclang.vendor.pygls.server import LanguageServer

import lsprotocol.types as lspt


class ALev(IntEnum):
    """Analysis Level successfully completed."""

    QUICK = 1
    DEEP = 2
    TYPE = 3


class ModuleInfo:
    """Module IR and Stats."""

    def __init__(
        self,
        ir: ast.Module,
        errors: list[Alert],
        warnings: list[Alert],
        alev: ALev,
        parent: Optional[ModuleInfo] = None,
    ) -> None:
        """Initialize module info."""
        self.ir = ir
        self.errors = errors
        self.warnings = warnings
        self.alev = alev
        self.parent: Optional[ModuleInfo] = parent
        self.diagnostics = self.gen_diagnostics()
        self.sem_tokens: list[int] = self.gen_sem_tokens()

    @property
    def uri(self) -> str:
        """Return uri."""
        return uris.from_fs_path(self.ir.loc.mod_path)

    @property
    def has_syntax_error(self) -> bool:
        """Return if there are syntax errors."""
        return len(self.errors) > 0 and self.alev == ALev.QUICK

    def update_with(self, new_info: ModuleInfo, refresh: bool = False) -> None:
        """Update module info."""
        self.ir = new_info.ir
        if refresh:
            self.errors = new_info.errors
            self.warnings = new_info.warnings
        else:
            self.errors += [i for i in new_info.errors if i not in self.errors]
            self.warnings += [i for i in new_info.warnings if i not in self.warnings]
        self.alev = new_info.alev
        self.diagnostics = self.gen_diagnostics()
        if self.alev == ALev.TYPE:
            self.sem_tokens = self.gen_sem_tokens()

    def gen_diagnostics(self) -> list[lspt.Diagnostic]:
        """Return diagnostics."""
        return [
            lspt.Diagnostic(
                range=create_range(error.loc),
                message=error.msg,
                severity=lspt.DiagnosticSeverity.Error,
            )
            for error in self.errors
        ] + [
            lspt.Diagnostic(
                range=create_range(warning.loc),
                message=warning.msg,
                severity=lspt.DiagnosticSeverity.Warning,
            )
            for warning in self.warnings
        ]

    def gen_sem_tokens(self) -> list[int]:
        """Return semantic tokens."""
        tokens = []
        prev_line, prev_col = 0, 0
        for node in self.ir._in_mod_nodes:
            if isinstance(node, ast.NameAtom) and node.sem_token:
                line, col_start, col_end = (
                    node.loc.first_line - 1,
                    node.loc.col_start - 1,
                    node.loc.col_end - 1,
                )
                length = col_end - col_start
                tokens += [
                    line - prev_line,
                    col_start if line != prev_line else col_start - prev_col,
                    length,
                    *node.sem_token,
                ]
                prev_line, prev_col = line, col_start
        return tokens


class JacLangServer(LanguageServer):
    """Class for managing workspace."""

    def __init__(self) -> None:
        """Initialize workspace."""
        super().__init__("jac-lsp", "v0.1")
        self.modules: dict[str, ModuleInfo] = {}
        self.executor = ThreadPoolExecutor()

    async def push_diagnostics(self, file_path: str) -> None:
        """Push diagnostics for a file."""
        if file_path in self.modules:
            self.publish_diagnostics(
                file_path,
                self.modules[file_path].diagnostics,
            )

    def unwind_to_parent(self, file_path: str) -> str:
        """Unwind to parent."""
        orig_file_path = file_path
        if file_path in self.modules:
            while cur := self.modules[file_path].parent:
                file_path = cur.uri
            if file_path == orig_file_path and (
                discover := self.modules[file_path].ir.annexable_by
            ):
                file_path = uris.from_fs_path(discover)
                self.quick_check(file_path)
        return file_path

    def update_modules(
        self, file_path: str, build: Pass, alev: ALev, refresh: bool = False
    ) -> None:
        """Update modules."""
        if not isinstance(build.ir, ast.Module):
            self.log_error("Error with module build.")
            return
        new_mod = ModuleInfo(
            ir=build.ir,
            errors=[
                i
                for i in build.errors_had
                if i.loc.mod_path == uris.to_fs_path(file_path)
            ],
            warnings=[
                i
                for i in build.warnings_had
                if i.loc.mod_path == uris.to_fs_path(file_path)
            ],
            alev=alev,
        )
        if file_path in self.modules:
            self.modules[file_path].update_with(new_mod, refresh=refresh)
        else:
            self.modules[file_path] = new_mod
        for p in build.ir.mod_deps.keys():
            uri = uris.from_fs_path(p)
            new_mod = ModuleInfo(
                ir=build.ir.mod_deps[p],
                errors=[i for i in build.errors_had if i.loc.mod_path == p],
                warnings=[i for i in build.warnings_had if i.loc.mod_path == p],
                alev=alev,
            )
            if not refresh and uri in self.modules:
                self.modules[uri].update_with(new_mod)
            else:
                self.modules[uri] = new_mod
            self.modules[uri].parent = (
                self.modules[file_path] if file_path != uri else None
            )

    def quick_check(self, file_path: str) -> bool:
        """Rebuild a file."""
        try:
            document = self.workspace.get_text_document(file_path)
            build = jac_str_to_pass(
                jac_str=document.source, file_path=document.path, schedule=[]
            )
        except Exception as e:
            self.log_error(f"Error during syntax check: {e}")
            return False
        self.update_modules(file_path, build, ALev.QUICK, refresh=True)
        return len(self.modules[file_path].errors) == 0

    def deep_check(self, file_path: str) -> bool:
        """Rebuild a file and its dependencies."""
        if file_path not in self.modules:
            self.quick_check(file_path)
        try:
            file_path = self.unwind_to_parent(file_path)
            build = jac_ir_to_pass(ir=self.modules[file_path].ir)
        except Exception as e:
            self.log_error(f"Error during syntax check: {e}")
            return False
        self.update_modules(file_path, build, ALev.DEEP)
        return len(self.modules[file_path].errors) == 0

    def type_check(self, file_path: str) -> bool:
        """Rebuild a file and its dependencies."""
        if file_path not in self.modules:
            self.deep_check(file_path)
        try:
            file_path = self.unwind_to_parent(file_path)
            build = jac_ir_to_pass(
                ir=self.modules[file_path].ir, schedule=type_checker_sched
            )
        except Exception as e:
            self.log_error(f"Error during type check: {e}")
            return False
        self.update_modules(file_path, build, ALev.TYPE)
        return len(self.modules[file_path].errors) == 0

    async def analyze_and_publish(self, uri: str, level: int = 2) -> None:
        """Analyze and publish diagnostics."""
        self.log_py(f"Analyzing {uri}...")
        success = await asyncio.get_event_loop().run_in_executor(
            self.executor, self.quick_check, uri
        )
        await self.push_diagnostics(uri)
        if success and level > 0:
            success = await asyncio.get_event_loop().run_in_executor(
                self.executor, self.deep_check, uri
            )
            await self.push_diagnostics(uri)
            if level > 1:
                await asyncio.get_event_loop().run_in_executor(
                    self.executor, self.type_check, uri
                )
                await self.push_diagnostics(uri)

    def get_completion(
        self, file_path: str, position: lspt.Position
    ) -> lspt.CompletionList:
        """Return completion for a file."""
        items = []
        document = self.workspace.get_text_document(file_path)
        current_line = document.lines[position.line].strip()
        if current_line.endswith("hello."):

            items = [
                lspt.CompletionItem(label="world"),
                lspt.CompletionItem(label="friend"),
            ]
        return lspt.CompletionList(is_incomplete=False, items=items)

    def rename_module(self, old_path: str, new_path: str) -> None:
        """Rename module."""
        if old_path in self.modules and new_path != old_path:
            self.modules[new_path] = self.modules[old_path]
            del self.modules[old_path]

    def delete_module(self, uri: str) -> None:
        """Delete module."""
        if uri in self.modules:
            del self.modules[uri]

    def formatted_jac(self, file_path: str) -> list[lspt.TextEdit]:
        """Return formatted jac."""
        try:
            document = self.workspace.get_text_document(file_path)
            format = jac_str_to_pass(
                jac_str=document.source,
                file_path=document.path,
                target=JacFormatPass,
                schedule=[FuseCommentsPass, JacFormatPass],
            )
            formatted_text = (
                format.ir.gen.jac
                if JacParser not in [e.from_pass for e in format.errors_had]
                else document.source
            )
        except Exception as e:
            self.log_error(f"Error during formatting: {e}")
            formatted_text = document.source
        return [
            lspt.TextEdit(
                range=lspt.Range(
                    start=lspt.Position(line=0, character=0),
                    end=lspt.Position(
                        line=len(document.source.splitlines()) + 1, character=0
                    ),
                ),
                new_text=(formatted_text),
            )
        ]

    def get_hover_info(
        self, file_path: str, position: lspt.Position
    ) -> Optional[lspt.Hover]:
        """Return hover information for a file."""
        node_selected = find_deepest_symbol_node_at_pos(
            self.modules[file_path].ir, position.line, position.character
        )
        value = self.get_node_info(node_selected) if node_selected else None
        if value:
            return lspt.Hover(
                contents=lspt.MarkupContent(
                    kind=lspt.MarkupKind.PlainText, value=f"{value}"
                ),
            )
        return None

    def get_node_info(self, node: ast.AstSymbolNode) -> Optional[str]:
        """Extract meaningful information from the AST node."""
        try:
            if isinstance(node, ast.NameAtom):
                node = node.name_of
            access = node.sym.access.value + " " if node.sym else None
            node_info = (
                f"({access if access else ''}{node.sym_category.value}) {node.sym_name}"
            )
            if node.name_spec.clean_type:
                node_info += f": {node.name_spec.clean_type}"
            if isinstance(node, ast.AstSemStrNode) and node.semstr:
                node_info += f"\n{node.semstr.value}"
            if isinstance(node, ast.AstDocNode) and node.doc:
                node_info += f"\n{node.doc.value}"
            if isinstance(node, ast.Ability) and node.signature:
                node_info += f"\n{node.signature.unparse()}"
            self.log_py(f"mypy_node: {node.gen.mypy_ast}")
        except AttributeError as e:
            self.log_warning(f"Attribute error when accessing node attributes: {e}")
        return node_info.strip()

    def get_document_symbols(self, file_path: str) -> list[lspt.DocumentSymbol]:
        """Return document symbols for a file."""
        if root_node := self.modules[file_path].ir._sym_tab:
            return collect_symbols(root_node)
        return []

    def get_definition(
        self, file_path: str, position: lspt.Position
    ) -> Optional[lspt.Location]:
        """Return definition location for a file."""
        node_selected: Optional[ast.AstSymbolNode] = find_deepest_symbol_node_at_pos(
            self.modules[file_path].ir, position.line, position.character
        )
        if node_selected:
            if (
                isinstance(node_selected, ast.Name)
                and node_selected.parent
                and isinstance(node_selected.parent, ast.ModulePath)
            ):
                spec = get_mod_path(node_selected.parent, node_selected)
                if spec:
                    return lspt.Location(
                        uri=uris.from_fs_path(spec),
                        range=lspt.Range(
                            start=lspt.Position(line=0, character=0),
                            end=lspt.Position(line=0, character=0),
                        ),
                    )
                else:
                    return None
            elif node_selected.parent and isinstance(
                node_selected.parent, ast.ModuleItem
            ):
                path_range = get_item_path(node_selected.parent)
                if path_range:
                    path, range = path_range
                    if path and range:
                        return lspt.Location(
                            uri=uris.from_fs_path(path),
                            range=lspt.Range(
                                start=lspt.Position(line=range[0], character=0),
                                end=lspt.Position(line=range[1], character=5),
                            ),
                        )
                else:
                    return None
            elif isinstance(node_selected, (ast.ElementStmt, ast.BuiltinType)):
                return None
            decl_node = (
                node_selected.parent.body.target
                if node_selected.parent
                and isinstance(node_selected.parent, ast.AstImplNeedingNode)
                and isinstance(node_selected.parent.body, ast.AstImplOnlyNode)
                else (
                    node_selected.sym.decl
                    if (node_selected.sym and node_selected.sym.decl)
                    else node_selected
                )
            )
            self.log_py(f"{node_selected}, {decl_node}")
            decl_uri = uris.from_fs_path(decl_node.loc.mod_path)
            try:
                decl_range = create_range(decl_node.loc)
            except ValueError:  # 'print' name has decl in 0,0,0,0
                return None
            decl_location = lspt.Location(
                uri=decl_uri,
                range=decl_range,
            )

            return decl_location
        else:
            return None

    def get_semantic_tokens(self, file_path: str) -> lspt.SemanticTokens:
        """Return semantic tokens for a file."""
        if file_path not in self.modules:
            return lspt.SemanticTokens(data=[])
        return lspt.SemanticTokens(data=self.modules[file_path].sem_tokens)

    def log_error(self, message: str) -> None:
        """Log an error message."""
        self.show_message_log(message, lspt.MessageType.Error)
        self.show_message(message, lspt.MessageType.Error)

    def log_warning(self, message: str) -> None:
        """Log a warning message."""
        self.show_message_log(message, lspt.MessageType.Warning)
        self.show_message(message, lspt.MessageType.Warning)

    def log_info(self, message: str) -> None:
        """Log an info message."""
        self.show_message_log(message, lspt.MessageType.Info)
        self.show_message(message, lspt.MessageType.Info)

    def log_py(self, message: str) -> None:
        """Log a message."""
        logging.info(message)



"""Ast build pass for Jaseci Ast.

This pass adds a more complete set of symbols from the AST to the
symbol table. This includes assignments, parameters, arch ref chains,
and more. This pass also links the symbols in the AST to their corresponding
sybmols in the symbol table (including uses).
"""

import ast as ast3

import jaclang.compiler.absyntree as ast
from jaclang.compiler.constant import Tokens as Tok
from jaclang.compiler.passes import Pass


class DefUsePass(Pass):
    """Jac Ast build pass."""

    def after_pass(self) -> None:
        """After pass."""

    def enter_architype(self, node: ast.Architype) -> None:
        """Sub objects.

        name: Name,
        arch_type: Token,
        access: Optional[SubTag[Token]],
        base_classes: Optional[SubNodeList[Expr]],
        body: Optional[SubNodeList[ArchBlockStmt] | ArchDef],
        doc: Optional[String] = None,
        semstr: Optional[String] = None,
        decorators: Optional[SubNodeList[Expr]] = None,
        """
        node.sym_tab.inherit_baseclasses_sym(node)

        def inform_from_walker(node: ast.AstNode) -> None:
            for i in (
                node.get_all_sub_nodes(ast.VisitStmt)
                + node.get_all_sub_nodes(ast.IgnoreStmt)
                + node.get_all_sub_nodes(ast.DisengageStmt)
                + node.get_all_sub_nodes(ast.EdgeOpRef)
            ):
                i.from_walker = True

        if node.arch_type.name == Tok.KW_WALKER:
            inform_from_walker(node)
            for i in self.get_all_sub_nodes(node, ast.Ability):
                if isinstance(i.body, ast.AbilityDef):
                    inform_from_walker(i.body)

    def enter_enum(self, node: ast.Enum) -> None:
        """Sub objects.

        name: Name,
        access: Optional[SubTag[Token]],
        base_classes: Optional[SubNodeList[Expr]],
        body: Optional[SubNodeList[EnumBlockStmt] | EnumDef],
        doc: Optional[String] = None,
        semstr: Optional[String] = None,
        decorators: Optional[SubNodeList[Expr]] = None,
        """
        node.sym_tab.inherit_baseclasses_sym(node)

    def enter_arch_ref(self, node: ast.ArchRef) -> None:
        """Sub objects.

        name_ref: NameType,
        arch: Token,
        """
        node.sym_tab.use_lookup(node)

    def enter_arch_ref_chain(self, node: ast.ArchRefChain) -> None:
        """Sub objects.

        archs: list[ArchRef],
        """
        node.sym_tab.chain_use_lookup(node.archs)

    def enter_param_var(self, node: ast.ParamVar) -> None:
        """Sub objects.

        name: Name,
        unpack: Optional[Token],
        type_tag: SubTag[ExprType],
        value: Optional[ExprType],
        """
        node.sym_tab.def_insert(node)

    def enter_has_var(self, node: ast.HasVar) -> None:
        """Sub objects.

        name: Name,
        type_tag: SubTag[SubNodeList[TypeSpec]],
        value: Optional[ExprType],
        """
        if isinstance(node.parent, ast.SubNodeList) and isinstance(
            node.parent.parent, ast.ArchHas
        ):
            node.sym_tab.def_insert(
                node,
                single_decl="has var",
                access_spec=node.parent.parent,
            )
        else:
            self.ice("Inconsistency in AST, has var should be under arch has")

    def enter_assignment(self, node: ast.Assignment) -> None:
        """Sub objects.

        target: SubNodeList[AtomType],
        value: Optional[ExprType | YieldStmt],
        type_tag: Optional[SubTag[ExprType]],
        is_static: bool = False,
        mutable: bool = True,
        """
        for i in node.target.items:
            if isinstance(i, ast.AtomTrailer):
                i.sym_tab.chain_def_insert(i.as_attr_list)
            elif isinstance(i, ast.AstSymbolNode):
                i.sym_tab.def_insert(i)
            else:
                self.error("Assignment target not valid")

    def enter_inner_compr(self, node: ast.InnerCompr) -> None:
        """Sub objects.

        is_async: bool,
        target: ExprType,
        collection: ExprType,
        conditional: Optional[ExprType],
        """
        if isinstance(node.target, ast.AtomTrailer):
            node.target.sym_tab.chain_def_insert(node.target.as_attr_list)
        elif isinstance(node.target, ast.AstSymbolNode):
            node.target.sym_tab.def_insert(node.target)
        else:
            self.error("Named target not valid")

    def enter_atom_trailer(self, node: ast.AtomTrailer) -> None:
        """Sub objects.

        target: AtomType,
        right: AtomType,
        is_scope_contained: bool,
        """
        chain = node.as_attr_list
        node.sym_tab.chain_use_lookup(chain)

    def enter_func_call(self, node: ast.FuncCall) -> None:
        """Sub objects.

        target: AtomType,
        params: Optional[SubNodeList[ExprType | Assignment]],
        """

    def enter_index_slice(self, node: ast.IndexSlice) -> None:
        """Sub objects.

        start: Optional[ExprType],
        stop: Optional[ExprType],
        is_range: bool,
        """

    def enter_special_var_ref(self, node: ast.SpecialVarRef) -> None:
        """Sub objects.

        var: Token,
        """
        node.sym_tab.use_lookup(node)

    def enter_edge_op_ref(self, node: ast.EdgeOpRef) -> None:
        """Sub objects.

        filter_type: Optional[ExprType],
        filter_cond: Optional[SubNodeList[BinaryExpr]],
        edge_dir: EdgeDir,
        from_walker: bool,
        """

    def enter_disconnect_op(self, node: ast.DisconnectOp) -> None:
        """Sub objects.

        edge_spec: EdgeOpRef,
        """

    def enter_connect_op(self, node: ast.ConnectOp) -> None:
        """Sub objects.

        conn_type: Optional[ExprType],
        conn_assign: Optional[SubNodeList[Assignment]],
        edge_dir: EdgeDir,
        """

    def enter_filter_compr(self, node: ast.FilterCompr) -> None:
        """Sub objects.

        compares: SubNodeList[BinaryExpr],
        """

    def enter_token(self, node: ast.Token) -> None:
        """Sub objects.

        name: str,
        value: str,
        col_start: int,
        col_end: int,
        pos_start: int,
        pos_end: int,
        """

    def enter_float(self, node: ast.Float) -> None:
        """Sub objects.

        name: str,
        value: str,
        line: int,
        col_start: int,
        col_end: int,
        pos_start: int,
        pos_end: int,
        """
        node.sym_tab.use_lookup(node)

    def enter_int(self, node: ast.Int) -> None:
        """Sub objects.

        name: str,
        value: str,
        line: int,
        col_start: int,
        col_end: int,
        pos_start: int,
        pos_end: int,
        """
        node.sym_tab.use_lookup(node)

    def enter_string(self, node: ast.String) -> None:
        """Sub objects.

        name: str,
        value: str,
        line: int,
        col_start: int,
        col_end: int,
        pos_start: int,
        pos_end: int,
        """
        node.sym_tab.use_lookup(node)

    def enter_bool(self, node: ast.Bool) -> None:
        """Sub objects.

        name: str,
        value: str,
        line: int,
        col_start: int,
        col_end: int,
        pos_start: int,
        pos_end: int,
        """
        node.sym_tab.use_lookup(node)

    def enter_builtin_type(self, node: ast.BuiltinType) -> None:
        """Sub objects.

        name: str,
        value: str,
        line: int,
        col_start: int,
        col_end: int,
        pos_start: int,
        pos_end: int,
        typ: type,
        """
        node.sym_tab.use_lookup(node)

    def enter_name(self, node: ast.Name) -> None:
        """Sub objects.

        name: str,
        value: str,
        col_start: int,
        col_end: int,
        pos_start: int,
        pos_end: int,
        """
        if not isinstance(node.parent, ast.AtomTrailer):
            node.sym_tab.use_lookup(node)

    def enter_in_for_stmt(self, node: ast.InForStmt) -> None:
        """Sub objects.

        target: ExprType,
        is_async: bool,
        collection: ExprType,
        body: SubNodeList[CodeBlockStmt],
        else_body: Optional[ElseStmt],
        """
        if isinstance(node.target, ast.AtomTrailer):
            node.target.sym_tab.chain_def_insert(node.target.as_attr_list)
        elif isinstance(node.target, ast.AstSymbolNode):
            node.target.sym_tab.def_insert(node.target)
        else:
            self.error("For loop assignment target not valid")

    def enter_delete_stmt(self, node: ast.DeleteStmt) -> None:
        """Sub objects.

        target: expression,
        """
        items = (
            node.target.values.items
            if isinstance(node.target, ast.TupleVal) and node.target.values
            else [node.target]
        )
        for i in items:
            if isinstance(i, ast.AtomTrailer):
                i.as_attr_list[-1].name_spec.py_ctx_func = ast3.Del
            elif isinstance(i, ast.AstSymbolNode):
                i.name_spec.py_ctx_func = ast3.Del
            else:
                self.error("Delete target not valid")

    def enter_expr_as_item(self, node: ast.ExprAsItem) -> None:
        """Sub objects.

        expr: ExprType,
        alias: Optional[ExprType],
        """
        if node.alias:
            if isinstance(node.alias, ast.AtomTrailer):
                node.alias.sym_tab.chain_def_insert(node.alias.as_attr_list)
            elif isinstance(node.alias, ast.AstSymbolNode):
                node.alias.sym_tab.def_insert(node.alias)
            else:
                self.error("For expr as target not valid")



"""Test ast build pass module."""

import os

import jaclang
from jaclang.utils.helpers import extract_headings, heading_to_snake
from jaclang.utils.lang_tools import AstTool
from jaclang.utils.test import TestCase


class JacFormatPassTests(TestCase):
    """Test pass module."""

    def setUp(self) -> None:
        """Set up test."""
        return super().setUp()

    def test_pass_template(self) -> None:
        """Basic test for pass."""
        out = AstTool().pass_template()
        self.assertIn("target: Expr,", out)
        self.assertIn("self, node: ast.ReturnStmt", out)
        self.assertIn("exprs: SubNodeList[ExprAsItem],", out)
        self.assertIn("value: str,", out)
        self.assertIn("def exit_module(self, node: ast.Module)", out)
        self.assertGreater(out.count("def exit_"), 20)

    def test_gendotfile(self) -> None:
        """Testing for HTML entity."""
        jac_file_path = os.path.join(
            os.path.dirname(jaclang.__file__), "../examples/reference/edges_walk.jac"
        )
        out = AstTool().ir(["ast.", jac_file_path])
        forbidden_strings = ["<<", ">>", "init", "super"]
        for i in forbidden_strings:
            self.assertNotIn(i, out)

    def test_print(self) -> None:
        """Testing for print AstTool."""
        jac_file = os.path.join(
            os.path.dirname(jaclang.__file__),
            "../examples/reference/names_and_references.jac",
        )
        msg = "error in " + jac_file
        out = AstTool().ir(["ast", jac_file])
        self.assertIn("+-- Token", out, msg)
        self.assertIsNotNone(out, msg=msg)

    def test_print_py(self) -> None:
        """Testing for print_py AstTool."""
        jac_py_directory = os.path.join(
            os.path.dirname(jaclang.__file__), "../examples/reference/"
        )
        jac_py_files = [
            f
            for f in os.listdir(jac_py_directory)
            if f.endswith(("names_and_references.jac", "names_and_references.py"))
        ]

        for file in jac_py_files:
            msg = "error in " + file
            out = AstTool().ir(["pyast", jac_py_directory + file])
            if file.endswith(".jac"):
                self.assertIn("Module(", out, msg)
                self.assertIsNotNone(out, msg=msg)
            elif file.endswith(".py"):
                if len(out.splitlines()) <= 4:
                    continue
                self.assertIn("Module(", out, msg)
                self.assertIsNotNone(out, msg=msg)

    def test_automated(self) -> None:
        """Testing for py, jac, md files for each content in Jac Grammer."""
        lark_path = os.path.join(os.path.dirname(jaclang.__file__), "compiler/jac.lark")
        headings_ = extract_headings(lark_path)
        snake_case_headings = [heading_to_snake(key) for key in headings_.keys()]
        refr_path = os.path.join(
            os.path.dirname(jaclang.__file__), "../examples/reference"
        )
        file_extensions = [".py", ".jac", ".md"]
        created_files = [f"{os.path.join(refr_path, 'introduction.md')}"]
        for heading_name in snake_case_headings:
            for extension in file_extensions:
                file_name = heading_name + extension
                file_path = os.path.join(refr_path, file_name)
                self.assertTrue(
                    os.path.exists(file_path), f"File '{file_path}' does not exist."
                )
                created_files.append(file_path)
        all_reference_files = [
            os.path.join(refr_path, file)
            for file in os.listdir(refr_path)
            if os.path.isfile(os.path.join(refr_path, file))
        ]
        other_reference_files = [
            os.path.basename(file)
            for file in all_reference_files
            if file not in created_files
        ]
        self.assertEqual(len(other_reference_files), 0)

    def test_py_jac_mode(self) -> None:
        """Testing for py_jac_mode support."""
        file = self.fixture_abs_path("../../../tests/fixtures/pyfunc.py")
        out = AstTool().ir(["unparse", file])
        self.assertIn("can my_print(x: object) -> None", out)

    def test_sym_sym_dot(self) -> None:
        """Testing for sym, sym. AstTool."""
        jac_file = os.path.join(
            os.path.dirname(jaclang.__file__), "../examples/reference/atom.jac"
        )
        out = AstTool().ir(["sym", jac_file])
        self.assertNotIn(
            "\n|   +-- ConnectionAbortedError\n|   |   +-- public var\n|   +-- ConnectionError\n|",
            out,
        )
        self.assertTrue(
            out.startswith("SymTable::Module(atom)\n+-- Symbols\n|   +-- (e)x\n|")
        )
        out = AstTool().ir(["sym.", jac_file])
        self.assertEqual('2 [label="(e)x"];', out.split("\n")[4])
        self.assertNotIn('[label="str"];', out)
